#!/usr/bin/python

###############################################################################
# This is the Ubuntu manpage repository generator and interface.
#
# Copyright (C) 2013 Dustin Kirkland <dustin.kirkland@gmail.com>
#
# This code was originally written by Dustin Kirkland <kirkland@ubuntu.com>,
# based on a framework by Kees Cook <kees@ubuntu.com>.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# On Debian-based systems, the complete text of the GNU General Public
# License can be found in /usr/share/common-licenses/GPL-3
###############################################################################

import argparse
import datetime
import errno
import gzip
import json
import logging
import os
import pprint
import re
import requests
import shutil
import signal
import subprocess
import sys
import tempfile

CONFIG_FILE = "%s/config.json" % os.path.dirname(os.path.realpath(__file__))
CONFIG = json.loads(open(CONFIG_FILE).read())
TEMPFILE = ""
TEMPDIR = ""
PACKAGES = ""
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)
logging.getLogger("requests").setLevel(logging.WARNING)
ENV = os.environ
ENV["W3MMAN_MAN"] = "man --no-hyphenation"
ENV["MAN_KEEP_FORMATTING"] = "1"
ENV["MANWIDTH"] = "120"

def signal_handler(signal, frame):
        print 'You pressed Ctrl+C!'
	cleanup()
        sys.exit(0)


def rm_rf(path):
	try:
		shutil.rmtree(path)
		return True
	except:
		# Handle freakish file permissions that might cause rmtree to fail
		proc = subprocess.Popen(["chmod", "-R", "700", path], stdout=subprocess.PIPE)
		output, _ = proc.communicate(None)
		proc = subprocess.Popen(["rm", "-rf", path], stdout=subprocess.PIPE)
		output, _ = proc.communicate(None)
		return False


def rm_f(path):
	try:
		if os.path.isfile(path) or os.path.islink(path):
			os.unlink(path)
		elif os.path.isdir(path):
			os.rmdir(path)
		return True
	except:
		logging.error("Could not remove [%s]" % path)
		return False


def symlink(target, link_name):
	rm_f(link_name)
	os.symlink(target, link_name)
	return True


def cleanup():
	for i in [TEMPFILE, TEMPDIR, PACKAGES]:
		logging.info("Removing [%s]" % i)
		rm_rf(i)


def error(msg):
	logging.error(msg)
	cleanup()
	os._exit(1)


def mkdir_p(path):
	try:
		os.makedirs(path)
	except OSError as exc:
		if exc.errno == errno.EEXIST and os.path.isdir(path):
			pass
		else:
			raise


def handle_deb(dist, deb, md5, forced):
	logging.info("Examining package [%s]" % deb)
	if forced or package_updated(deb, md5):
		logging.info("  Downloading new package [%s]" % deb)
		if fetch_manpages(dist, deb, md5):
			return True
		else:
			return False
	else:
		logging.info("  Skipping unchanged package [%s]" % deb)
		return True


def package_updated(deb, md5):
	cache = "%s/manpages/.cache/%s" % (CONFIG["PUBLIC_HTML_DIR"], deb)
	if os.path.exists(cache) and open(cache).read() == md5:
		return False
	else:
		return True


def get_manpage_paths(deb):
	manpages = []
	proc = subprocess.Popen(["dpkg-deb", "-c", deb], stdout=subprocess.PIPE)
	output, _ = proc.communicate(None)
	for line in output.splitlines():
		f = line.split()[5]
		if f[0:16] == "./usr/share/man/" and f[-3:] == ".gz":
			# TODO: handle postgres' weird location usr/share.*/man/
			manpages.append(f[2:])
	return manpages


def get_source_package(deb):
	# We want the source package name, but also handle packages which are the same name as their source
	proc = subprocess.Popen(["dpkg-deb", "-f", deb, "Package", "Source"], stdout=subprocess.PIPE)
	output, _ = proc.communicate(None)
	try:
		lines = output.splitlines()
		return lines[len(lines)-1].split(": ")[1]
	except:
		return ""
	return ""


def fetch_manpages(dist, deb, md5):
	url = "%s/%s" % (CONFIG["WEBARCHIVE"], deb)
	logging.info("    Fetching url [%s]" % url)
	if download_file(url, TEMPFILE):
		logging.info("    Looking for manpages in [%s] [%s]" % (dist, deb))
		manpages = get_manpage_paths(TEMPFILE)
		if manpages:
			logging.info("    Found manpages %s" % manpages)
			# Extract files to a temporary directory
			proc = subprocess.Popen(['dpkg-deb', '-x', TEMPFILE, TEMPDIR], stdout=subprocess.PIPE)
			out, _ = proc.communicate(None)
			html_dir = "%s/manpages/%s" % (CONFIG["PUBLIC_HTML_DIR"], dist)
			gz_dir = "%s/manpages.gz/%s" % (CONFIG["PUBLIC_HTML_DIR"], dist)
			for i in manpages:
				target = "%s/%s" % (TEMPDIR, i)
				if os.path.isfile(target) and os.path.getsize(target) == 0:
					# Skip empty manpages as quickly as possible
					logging.info("      Skipping empty manpage [%s]" % i)
					continue
				short_name = re.sub("^.*usr\/share.*\/man\/", "", i)
				short_name = re.sub("\.gz$", "", short_name)
				logging.info("      Considering entry [%s]" % short_name)
				html_out = "%s/%s.html" % (html_dir, short_name)
				gz_out = "%s/%s.gz" % (gz_dir, short_name)
				mkdir_p(os.path.dirname(html_out))
				mkdir_p(os.path.dirname(gz_out))
				if os.path.islink(target):
					dest = re.sub(r"\.gz$", ".html", os.readlink(target))
					if symlink(dest, html_out):
						logging.info("        Successfully created symlink [%s] -> [%s]" % (dest, html_out))
				else:
					html = render_html(target, dist, deb)
					if html:
						# Write out our post-processed HTML
						rm_f(html_out)
						with open(html_out, "w") as f:
							f.write(html)
							logging.info("        Successfully created HTML manpage [%s]" % html_out)
					else:
						# Empty output from w3m, check if we should symlink this?
						lines = gzip.open(target, "rb").read().splitlines()
						if len(lines) > 0:
							parts = lines[0].split()
							if len(parts) == 2 and parts[0] == ".so":
								if symlink("../%s.html" % parts[1], html_out):
									logging.info("        Successfully created symlink [%s] -> [%s]" % ("../%s.html" % parts[1], html_out))
						else:
							logging.info("        Empty manpage [%s]" % target)
				# Move the gzip'd manpage over to our collection
				shutil.move(target, gz_out)
				logging.info("        Successfully copied original manpage [%s]" % gz_out)
		else:
			logging.info("    No manpages found in [%s]" % deb)
		cache_file = "%s/manpages/.cache/%s" % (CONFIG["PUBLIC_HTML_DIR"], deb)
		mkdir_p(os.path.dirname(cache_file))
		rm_f(cache_file)
		with open(cache_file, "w") as f:
			f.write(md5)
	return False


def render_html(target, dist, deb):
	# Render manpage to HTML using w3m
	proc = subprocess.Popen(["/usr/lib/w3m/cgi-bin/w3mman2html.cgi", "local=%s" % target], env=ENV, stdout=subprocess.PIPE)
	output, _ = proc.communicate(None)
	if not output:
		return ""
	# Post process our HTML
	# Remove w3m's header and footer
	lines = output.splitlines()[7:-2]
	if len(lines) < 3:
		return ""
	# Grab the title, supressing any html markup
	title = re.sub(r'<[^>]+>', "", lines[2])
	binary_package = re.sub(r'^.*/(.*)\.deb$', r'\1', deb)
	package_link = "https://launchpad.net/ubuntu/%s/+package/%s" % (dist, binary_package)
	bug_link = "https://bugs.launchpad.net/ubuntu/+source/%s/+filebug-advanced" % get_source_package(TEMPFILE)
	html = []
	html.append("<!--#include virtual='/above1.html' -->\n")
	html.append("%s\n" % title)
	html.append("<!--#include virtual='/above2.html' -->\n")
	html.append("Provided by: <a href='%s'>%s</a> <a href='%s' title='Report a bug in the content of this documentation'><img src='/img/bug.png' alt='bug' border=0></a><br><pre>" % (str(package_link), str(binary_package), str(bug_link)))
	for l in lines:
		# Convert section titles (bold lines) to h4's
		l = re.sub(r'^<b>(.*)</b>$', r'</pre><h4><b>\1</b></h4><pre>', l)
		# Convert local file:/// references to relative url links
		l = re.sub(r'<a href="file:///[^\?]*\?([^\(]*)\(([^\)]*)\)">', r'<a href="../man\2/\1.\2.html">', l)
		html.append(l)
	html.append("</pre><!--#include virtual='/below.html' -->\n")
	return '\n'.join(html)


def link_en_locale(dist):
	mkdir_p("%s/manpages/%s/en" % (CONFIG["PUBLIC_HTML_DIR"], dist))
	for i in range(1, 10):
		for j in ["manpages", "manpages.gz"]:
			mkdir_p("%s/%s/%s/en" % (CONFIG["PUBLIC_HTML_DIR"], j, dist))
			d = "%s/%s/%s/en/man%d" % (CONFIG["PUBLIC_HTML_DIR"], j, dist, i)
			if os.path.islink(d):
				continue
			elif os.path.isdir(d):
				d_bak = "%s.bak" % d
				shutil.move(d, d_bak)
				os.symlink("../man%d" % i, d)
				for k in os.listdir(d_bak):
					shutil.move("%s/%s" % (d_bak, k), d)
				os.rmdir(d_bak)
			else:
				symlink("../man%d" % i, d)
	return True


def download_file(url, output_file):
	response = requests.get(url)
	if response.status_code == 200:
		rm_f(output_file)
		with open(output_file, "w") as f:
			for chunk in response.iter_content():
				f.write(chunk)
		return True
	else:
		return False


def open_sitemap(count):
	path = "%s/manpages/sitemap_%s.xml" % (CONFIG["PUBLIC_HTML_DIR"], count)
	sitemap = open(path, "w")
	sitemap.write("<?xml version='1.0' encoding='UTF-8'?>\n")
	sitemap.write("<urlset xmlns='http://www.sitemaps.org/schemas/sitemap/0.9'>\n")
	logging.info("Creating new sitemap [%s]" % path)
	return sitemap


def close_sitemap(sitemap):
	sitemap.write("</urlset>")
	sitemap.close()


def make_sitemaps():
	path = "%s/manpages" % CONFIG["PUBLIC_HTML_DIR"]
	logging.info("Making sitemaps of [%s]" % path)
	paths = []
	for root, subdirs, files in os.walk(path, followlinks=True):
		files = [os.path.join(root, f) for f in files]
		# Find all files that end in .html, but exclude the ones symlinked to our
		# English subdirectory
		paths.extend([f for f in files if f.endswith(".html") and f.find("/en/man") < 0])
	paths = sorted(set(paths))
	logging.info("  Sitemap URLs total [%s]" % len(paths))
	i = 0
	limit = 50000
	sitemap_count = 0
	sitemap = open_sitemap(sitemap_count)
	for p in paths:
		if i < limit:
			i += 1
		else:
			close_sitemap(sitemap)
			sitemap_count += 1
			sitemap = open_sitemap(sitemap_count)
			i = 1
		url = "%s/%s" % (CONFIG["SITE"], p.replace(CONFIG["PUBLIC_HTML_DIR"], ""))
		sitemap.write("<url><loc>%s</loc></url>\n" % url)
	close_sitemap(sitemap)
	lastmod = datetime.datetime.now().isoformat()
	path = "%s/manpages/sitemaps.xml" % CONFIG["PUBLIC_HTML_DIR"]
	logging.info("Writing sitemaps index [%s]" % path)
	rm_f(path)
	with open(path, "w") as f:
		f.write("<?xml version='1.0' encoding='UTF-8'?>\n")
		f.write("<sitemapindex xmlns='http://www.sitemaps.org/schemas/sitemap/0.9'>\n")
		for i in range(0, sitemap_count):
			f.write("  <sitemap>\n")
			f.write("    <loc>%s/manpages/sitemap_%s.xml</loc>\n" % (CONFIG["SITE"], i))
			f.write("    <lastmod>%s</lastmod>\n" % lastmod)
			f.write("  </sitemap>\n")
		f.write("</sitemapindex>\n")


if __name__ == '__main__':
	try:
		signal.signal(signal.SIGINT, signal_handler)
		mkdir_p("%s/manpages" % CONFIG["PUBLIC_HTML_DIR"])
		fd, TEMPFILE = tempfile.mkstemp(dir=CONFIG["TMPDIR"], prefix='manpages-pkg-', suffix='.deb')
		fd, PACKAGES = tempfile.mkstemp(dir=CONFIG["TMPDIR"], prefix='manpages-packages-', suffix='.gz')
		TEMPDIR = tempfile.mkdtemp(dir=CONFIG["TMPDIR"], prefix='manpages-dir-')
		logging.info("Created temporary workspace [%s, %s, %s]" % (TEMPFILE, PACKAGES, TEMPDIR))
		parser = argparse.ArgumentParser(description='Create a repository of manpages')
		parser.add_argument('-f', '--force', help='Ignore caching and force a full repository creation', action="store_true", default=False)
		parser.add_argument('-s', '--sitemaps', help='Make sitemaps for search engine indexing', action="store_true", default=True)
		parser.add_argument('-d', '--distros', help='Distributions to process, comma separated (eg: precise,lucid,hardy)')
		parser.add_argument('-r', '--repos', help='Repositories to process, comma separated (eg: main,universe)')
		parser.options = parser.parse_args()
		if parser.options.distros:
			distros = parser.options.distros.split(",")
		else:
			distros = CONFIG["DISTROS"]
		if parser.options.repos:
			repos = parser.options.repos.split(",")
		else:
			repos = CONFIG["REPOS"]
		for d in distros:
			mkdir_p("%s/manpages/.cache" % (CONFIG["PUBLIC_HTML_DIR"]))
			mkdir_p("%s/manpages.gz/%s" % (CONFIG["PUBLIC_HTML_DIR"], d))
			link_en_locale(d)
			for r in repos:
				url = "%s/dists/%s/%s/binary-i386/Packages.gz" % (CONFIG["WEBARCHIVE"], d, r)
				logging.info("Fetching package list [%s]" % url)
				download_file(url, PACKAGES)
				deb = False
				md5 = False
				for line in gzip.open(PACKAGES, "rb").read().splitlines():
					if line.startswith("Filename:") and line.endswith(".deb"):
						deb = line.split()[1]
					elif line.startswith("MD5sum:"):
						md5 = line.split()[1]
					if deb and md5:
						handle_deb(d, deb, md5, parser.options.force)
						deb = False
						md5 = False
		if parser.options.sitemaps:
			make_sitemaps()
	except (Exception,):
		cleanup()
		#e = sys.exc_info()[1]
		#error("%s" % (str(e)))
		raise
	cleanup()
	os._exit(0)
